import logging
import click
from ephor_cli.types import AgentConfig
import yaml
import os
import sys
import requests
from typing import Dict, Any, Optional
from dotenv import load_dotenv
from google_a2a.common.types import (
    AgentSkill as A2AAgentSkill,
    AgentCapabilities as A2AAgentCapabilities,
    AgentCard,
)
from ephor_cli.server import A2AProxyServer
from ephor_cli.service.server.server import ConversationServer
from fastapi import FastAPI, APIRouter
from fastapi.middleware.cors import CORSMiddleware
import colorama
from colorama import Fore, Style

# Define version
__version__ = "0.2.2"

load_dotenv()

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def check_required_env_vars(env_vars: list[str]) -> bool:
    """Check if required environment variables are set."""
    missing_vars = []
    for var in env_vars:
        if not os.environ.get(var):
            missing_vars.append(var)

    if missing_vars:
        click.echo(
            f"{Fore.RED}Error: Missing required environment variables: {', '.join(missing_vars)}{Style.RESET_ALL}"
        )
        click.echo(
            f"{Fore.YELLOW}Please set these variables in your environment or .env file.{Style.RESET_ALL}"
        )
        return False
    return True


def load_agent_config(config_path: str) -> AgentConfig:
    """Load agent configuration from a YAML file."""
    try:
        with open(config_path, "r") as file:
            config_dict = yaml.safe_load(file)
            return AgentConfig(**config_dict)
    except Exception as e:
        logger.error(f"Failed to load config from {config_path}: {e}")
        raise


def create_agent_card(
    config: AgentConfig, host: str, port: int
) -> tuple[AgentCard, str, Optional[Dict[str, Any]]]:
    """Create an AgentCard from a configuration dictionary."""
    agent_name = config.name
    agent_id = agent_name.lower().replace(" ", "-")

    # Create skills
    skills = []
    for skill_config in config.skills:
        skill = A2AAgentSkill(
            id=skill_config.id,
            name=skill_config.name,
            description=skill_config.description,
            tags=skill_config.tags,
            examples=skill_config.examples,
            inputModes=skill_config.inputModes,
            outputModes=skill_config.outputModes,
        )
        skills.append(skill)

    # Create capabilities
    capabilities = A2AAgentCapabilities(streaming=config.capabilities.streaming)

    # Create agent card
    agent_card = AgentCard(
        name=config.name,
        description=config.description,
        url=f"http://{host}:{port}/{agent_id}",
        version=config.version,
        defaultInputModes=["text"],
        defaultOutputModes=["text"],
        capabilities=capabilities,
        skills=skills,
    )

    return agent_card, agent_id


@click.group()
def cli():
    """Ephor CLI for managing A2A proxy agents."""
    pass


@cli.command()
@click.option("--host", default="localhost", help="Host to run the server on")
@click.option("--port", default=10002, help="Port to run the server on")
@click.option(
    "--api-url", default="http://localhost:3000/api", help="API URL for agent registry"
)
def start_agent_server(host, port, api_url):
    """Start an A2A proxy server without pre-registering any agents."""
    # Check for required environment variables
    if not check_required_env_vars(["EPHOR_API_KEY", "ANTHROPIC_API_KEY"]):
        sys.exit(1)

    # Create and start the server
    server = A2AProxyServer(host=host, port=port, api_url=api_url)

    logger.info(f"Starting server on http://{host}:{port}")
    click.echo(
        f"Server running at: {Fore.YELLOW}{Style.BRIGHT}http://{host}:{port}{Style.RESET_ALL}"
    )
    click.echo(
        f"Your agent URL will be {Fore.CYAN}http://{host}:{port}/<agent_id> {Style.RESET_ALL}"
    )
    server.start()


@cli.command()
@click.option(
    "--output", "-o", default=None, help="Output file path for the agent config"
)
def create_agent(output):
    """Create a new agent configuration file by answering prompts."""
    try:
        colorama.init()
    except ImportError:
        # Fallback if colorama is not installed
        class DummyFore:
            def __getattr__(self, _):
                return ""

        class DummyStyle:
            def __getattr__(self, _):
                return ""

        global Fore, Style
        Fore = DummyFore()
        Style = DummyStyle()

    click.echo(
        f"{Fore.GREEN}===== Creating New Agent Configuration ====={Style.RESET_ALL}"
    )

    # Collect basic information
    name = click.prompt(f"{Fore.CYAN}Agent name{Style.RESET_ALL}")
    description = click.prompt(f"{Fore.CYAN}Agent description{Style.RESET_ALL}")
    version = click.prompt(
        f"{Fore.CYAN}Agent version{Style.RESET_ALL}", default="1.0.0"
    )

    # Capabilities
    streaming = click.confirm(
        f"{Fore.CYAN}Enable streaming capability?{Style.RESET_ALL}", default=True
    )

    # Skills
    skills = []
    click.echo(f"{Fore.YELLOW}Now let's define the agent's skills:{Style.RESET_ALL}")

    while True:
        if skills and not click.confirm(
            f"{Fore.YELLOW}Add another skill?{Style.RESET_ALL}", default=False
        ):
            break

        skill_id = click.prompt(f"{Fore.CYAN}Skill ID{Style.RESET_ALL}")
        skill_name = click.prompt(f"{Fore.CYAN}Skill name{Style.RESET_ALL}")
        skill_description = click.prompt(
            f"{Fore.CYAN}Skill description{Style.RESET_ALL}"
        )

        # Tags
        tags = []
        click.echo(f"{Fore.BLUE}Enter tags (empty line to finish):{Style.RESET_ALL}")
        while True:
            tag = click.prompt(f"{Fore.BLUE}Tag{Style.RESET_ALL}", default="")
            if not tag:
                break
            tags.append(tag)

        # Examples
        examples = []
        click.echo(
            f"{Fore.BLUE}Enter example queries (empty line to finish):{Style.RESET_ALL}"
        )
        while True:
            example = click.prompt(f"{Fore.BLUE}Example{Style.RESET_ALL}", default="")
            if not example:
                break
            examples.append(example)

        skill = {
            "id": skill_id,
            "name": skill_name,
            "description": skill_description,
            "tags": tags,
            "examples": examples,
            "inputModes": ["text"],
            "outputModes": ["text"],
        }

        skills.append(skill)

    # Prompt
    click.echo(f"{Fore.YELLOW}Enter the system prompt for your agent:{Style.RESET_ALL}")
    prompt = click.edit(text="You are an agent. Your job is to...\n")

    # MCP Servers
    mcp_servers = []
    if click.confirm(
        f"{Fore.YELLOW}Would you like to configure MCP servers?{Style.RESET_ALL}",
        default=False,
    ):
        while True:
            server_name = click.prompt(f"{Fore.CYAN}Server name{Style.RESET_ALL}")
            server_url = click.prompt(f"{Fore.CYAN}Server URL{Style.RESET_ALL}")
            server_transport = click.prompt(
                f"{Fore.CYAN}Transport type{Style.RESET_ALL}",
                type=click.Choice(["sse", "websocket", "http"], case_sensitive=False),
                default="sse",
            )

            mcp_servers.append(
                {"name": server_name, "url": server_url, "transport": server_transport}
            )

            if not click.confirm(
                f"{Fore.YELLOW}Add another MCP server?{Style.RESET_ALL}", default=False
            ):
                break

    # Create the config dictionary
    config = {
        "name": name,
        "description": description,
        "version": version,
        "capabilities": {"streaming": streaming},
        "skills": skills,
        "prompt": prompt,
    }

    if mcp_servers:
        config["mcpServers"] = mcp_servers

    # Determine output file path
    if not output:
        output = f"agent-{name.lower().replace(' ', '-')}.yml"
        if not output.endswith(".yml"):
            output += ".yml"

    # Save to file
    output_path = os.path.abspath(output)
    with open(output_path, "w") as f:
        yaml.dump(config, f, default_flow_style=False, sort_keys=False)

    click.echo(
        f"{Fore.GREEN}Agent configuration saved to: {output_path}{Style.RESET_ALL}"
    )


@cli.command()
@click.option(
    "--api-url",
    default="http://localhost:3000/api",
    help="API URL for agent registry",
)
def list_agents(api_url):
    """List available agents from the API."""
    # Check for required environment variables
    if not check_required_env_vars(["EPHOR_API_KEY"]):
        sys.exit(1)

    click.echo(f"{Fore.GREEN}===== Available Agents ====={Style.RESET_ALL}")

    page_size = 10
    next_token = None
    has_more = True

    while has_more:
        try:
            headers = {"x-api-key": os.environ.get("EPHOR_API_KEY")}
            params = {"limit": page_size}

            if next_token:
                params["nextToken"] = next_token

            response = requests.get(f"{api_url}/agents", params=params, headers=headers)

            if response.status_code != 200:
                click.echo(
                    f"{Fore.RED}Error: Failed to fetch agents: {response.text}{Style.RESET_ALL}"
                )
                sys.exit(1)

            data = response.json()
            agents = data.get("agents", [])
            next_token = data.get("nextToken")

            if not agents:
                if next_token is None:
                    click.echo(f"{Fore.YELLOW}No agents found.{Style.RESET_ALL}")
                else:
                    click.echo(
                        f"{Fore.YELLOW}No more agents to display.{Style.RESET_ALL}"
                    )
                break

            for agent in agents:
                agent_name = agent.get("name", "Unknown")
                agent_description = agent.get("description", "No description available")
                agent_id = agent.get("id", "Unknown")
                agent_version = agent.get("version", "Unknown")

                click.echo(f"{Fore.CYAN}{agent_name} (ID: {agent_id}){Style.RESET_ALL}")
                click.echo(f"  Description: {agent_description}")
                click.echo(f"  Version: {agent_version}")
                click.echo("")

            has_more = next_token is not None

            if has_more:
                click.echo(
                    f"{Fore.YELLOW}Press Enter to show more agents or Ctrl+C to exit...{Style.RESET_ALL}"
                )
                try:
                    input()
                except KeyboardInterrupt:
                    click.echo(
                        f"{Fore.YELLOW}Operation canceled by user.{Style.RESET_ALL}"
                    )
                    break

        except requests.exceptions.RequestException as e:
            click.echo(f"{Fore.RED}Error communicating with API: {e}{Style.RESET_ALL}")
            sys.exit(1)


@cli.command()
@click.option("--host", default="0.0.0.0", help="Host to run the server on")
@click.option("--port", default=12000, help="Port to run the server on")
def start_conversation_server(host, port):
    """Start the Conversation Server for agent interactions."""
    # Check for required environment variables
    if not check_required_env_vars(["GOOGLE_API_KEY"]):
        sys.exit(1)

    click.echo(
        f"{Fore.GREEN}Starting Conversation Server on http://{host}:{port}{Style.RESET_ALL}"
    )

    # Create FastAPI app and router
    app = FastAPI()

    # Add CORS middleware to allow all origins
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    router = APIRouter()

    # Initialize the conversation server with the router
    ConversationServer(router)

    # Include the router in the app
    app.include_router(router)

    # Start the server
    import uvicorn

    uvicorn.run(
        app,
        host=host,
        port=port,
        timeout_graceful_shutdown=0,
    )


@cli.command()
@click.option("--config", "-c", required=True, help="Path to agent config file")
@click.option(
    "--api-url",
    default="http://localhost:3000/api",
    help="API URL for agent registry",
)
def push(config, api_url):
    """Push an agent configuration to the registry API."""

    # Check for required environment variables
    if not check_required_env_vars(["EPHOR_API_KEY"]):
        sys.exit(1)

    try:
        # Load the agent config
        logger.info(f"Loading agent config from {config}")
        agent_config = load_agent_config(config)

        # Convert agent config to dictionary for API
        agent_data = {
            "name": agent_config.name,
            "description": agent_config.description,
            "version": agent_config.version,
            "capabilities": {"streaming": agent_config.capabilities.streaming},
            "skills": [],
            "prompt": agent_config.prompt,
        }

        # Add skills
        for skill in agent_config.skills:
            skill_data = {
                "id": skill.id,
                "name": skill.name,
                "description": skill.description,
                "tags": skill.tags,
                "examples": skill.examples,
                "inputModes": skill.inputModes,
                "outputModes": skill.outputModes,
            }
            agent_data["skills"].append(skill_data)

        # Add MCP servers if present
        if agent_config.mcpServers:
            agent_data["mcpServers"] = []
            for server in agent_config.mcpServers:
                agent_data["mcpServers"].append(
                    {
                        "name": server.name,
                        "url": server.url,
                        "transport": server.transport,
                    }
                )

        # Check if agent already exists
        agent_name = agent_config.name
        try:
            headers = {"x-api-key": os.environ.get("EPHOR_API_KEY")}
            response = requests.get(
                f"{api_url}/agents?name={agent_name}", headers=headers
            )
            existing_agents = response.json()
            agent_exists = False
            for agent in existing_agents:
                if agent.get("name") == agent_name:
                    agent_exists = True
                    agent_id = agent.get("id")
                    break

            if agent_exists:
                click.echo(
                    f"{Fore.YELLOW}Agent '{agent_name}' already exists.{Style.RESET_ALL}"
                )
                update = click.confirm(
                    f"{Fore.CYAN}Do you want to update the existing agent?{Style.RESET_ALL}",
                    default=True,
                )

                if update:
                    # Update existing agent
                    response = requests.put(
                        f"{api_url}/agents/{agent_id}", json=agent_data, headers=headers
                    )
                    if response.status_code in [200, 201, 204]:
                        click.echo(
                            f"{Fore.GREEN}Agent '{agent_name}' updated successfully.{Style.RESET_ALL}"
                        )
                    else:
                        click.echo(
                            f"{Fore.RED}Failed to update agent: {response.text}{Style.RESET_ALL}"
                        )
                else:
                    # Create with new name
                    new_name = click.prompt(
                        f"{Fore.CYAN}Enter a new unique name for the agent{Style.RESET_ALL}"
                    )
                    agent_data["name"] = new_name
                    response = requests.post(
                        f"{api_url}/agents", json=agent_data, headers=headers
                    )
                    if response.status_code in [200, 201]:
                        click.echo(
                            f"{Fore.GREEN}Agent '{new_name}' created successfully.{Style.RESET_ALL}"
                        )
                    else:
                        click.echo(
                            f"{Fore.RED}Failed to create agent: {response.text}{Style.RESET_ALL}"
                        )
            else:
                # Create new agent
                response = requests.post(
                    f"{api_url}/agents", json=agent_data, headers=headers
                )
                if response.status_code in [200, 201]:
                    click.echo(
                        f"{Fore.GREEN}Agent '{agent_name}' created successfully.{Style.RESET_ALL}"
                    )
                else:
                    click.echo(
                        f"{Fore.RED}Failed to create agent: {response.text}{Style.RESET_ALL}"
                    )

        except requests.exceptions.RequestException as e:
            click.echo(f"{Fore.RED}Error communicating with API: {e}{Style.RESET_ALL}")
            sys.exit(1)

    except Exception as e:
        logger.error(f"Failed to push agent: {e}")
        click.echo(f"{Fore.RED}Error: Failed to push agent: {e}{Style.RESET_ALL}")
        sys.exit(1)


@cli.command()
def version():
    """Print the current version of the Ephor CLI."""
    click.echo(f"Ephor CLI version: {__version__}")


def main():
    cli()


if __name__ == "__main__":
    main()
